#include <glad/gl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "diffren.h"
#include "diffren_model.h"

/* Implements a deferred renderer for PBR models */

typedef struct {
    /* 16bit RGB float for view space positions */
    GLuint position;

    /* RGB8 Buffer for surface base color  */
    GLuint albedo;

    /* 16bit float RGBs for view space normals */
    GLuint normal;

    /* 16bit float RGB for occlusion, roughness and metallic */
    GLuint orm;

    /* D32 for depth */
    GLuint depth;

    /* Framebuffer object for the whole gbuffer */
    GLuint fbo;
} GBuffer;

// Must match gl std140 layout
typedef struct {
    M4 view;
    M4 projection;
} FrameMatrices;

#define GEOMETRY_PASS_FRAME_MATRICES_BINDING 0
#define GEOMETRY_PASS_MODEL_LOCATION 0
#define GEOMETRY_PASS_ALBEDO_LOCATION 1
#define GEOMETRY_PASS_NORMAL_LOCATION 2
#define GEOMETRY_PASS_ORM_LOCATION 3

typedef struct {
    /* Shader program for rendering to the gbuffer */
    GLuint prog;

    /* Buffer containing the view/projection matrices,
     * at binding = 0 */
    GLuint frame_matrices_ubo;
} GeometryPass;

typedef struct {
    /* 16bit RGB float for HDR lighting computations */
    GLuint color;

    /* Framebuffer object */
    GLuint fbo;
} HDRBuffer;

#define LIGHTING_PASS_LIGHTS_BINDING 0
#define LIGHTING_PASS_LIGHT_COUNT_LOCATION 0
#define LIGHTING_PASS_AMBIENT_COLOR_LOCATION 1
#define LIGHTING_PASS_POSITION_LOCATION 2
#define LIGHTING_PASS_ALBEDO_LOCATION 3
#define LIGHTING_PASS_NORMAL_LOCATION 4
#define LIGHTING_PASS_ORM_LOCATION 5

typedef struct {
    /* Shader program for rendering the final image */
    GLuint prog;

    GLuint lights_sbo;
} LightingPass;

#define TONE_MAPPING_PASS_COLOR_LOCATION 0

typedef struct {
    GLuint prog;
} ToneMappingPass;

#define PIXEL_CLASSIFIER_PASS_NORMAL_BINDING 0
#define PIXEL_CLASSIFIER_PASS_CLASSIFICATION_BINDING 1

typedef struct {
    GLuint prog;
} PixelClassifierPass;

typedef struct {
    // UI8 buffer for pixel classification
    GLuint classification;
} DifferentiationBuffer;

#define MAX_LIGHTS 128

// Must match gl std430 layout
typedef struct {
    V4 position_radius;
    V4 color;
} Light;

struct DeferredRenderer {
    GBuffer gbuf;
    GeometryPass geometry_pass;

    HDRBuffer hdr_buf;
    LightingPass lighting_pass;
    ToneMappingPass tone_mapping_pass;

    PixelClassifierPass classifier_pass;
    DifferentiationBuffer diff_buffer;

    /* "Empty" vao for the full-screen triangle */
    GLuint full_screen_triangle_vao;

    int width;
    int height;

    /* Per Frame Data */
    V3 eye_pos;
    V3 center;
    V3 up;
    M4 projection;
    V3 ambient_color;

    Light lights[MAX_LIGHTS];
    unsigned int light_count;
    int light_ssbo_needs_update;
};

static void CreateGBuffer(int w, int h, GBuffer *gbuf) {
    glGenFramebuffers(1, &gbuf->fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, gbuf->fbo);

    /* Create the texture objects in one batch */
    GLuint tex_objs[4];
    glGenTextures(4, tex_objs);
    gbuf->position = tex_objs[0];
    gbuf->albedo = tex_objs[1];
    gbuf->normal = tex_objs[2];
    gbuf->orm = tex_objs[3];

    /* position buffer */
    glBindTexture(GL_TEXTURE_2D, gbuf->position);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F, w, h, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, gbuf->position, 0);

    /* albedo buffer */
    glBindTexture(GL_TEXTURE_2D, gbuf->albedo);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, w, h, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, gbuf->albedo, 0);

    /* normal buffer */
    glBindTexture(GL_TEXTURE_2D, gbuf->normal);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F, w, h, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT2, GL_TEXTURE_2D, gbuf->normal, 0);

    /* occlusion/roughness/metallic buffer */
    glBindTexture(GL_TEXTURE_2D, gbuf->orm);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F, w, h, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT3, GL_TEXTURE_2D, gbuf->orm, 0);

    /* 32bit depth buffer */
    glGenRenderbuffers(1, &gbuf->depth);
    glBindRenderbuffer(GL_RENDERBUFFER, gbuf->depth);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT32F, w, h);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, gbuf->depth);

    GLuint attachments[4] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2, GL_COLOR_ATTACHMENT3 };
    glDrawBuffers(4, attachments);
    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        fprintf(stderr, "ERROR::FRAMEBUFFER:: Framebuffer is not complete!\n");
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

/* Return a zero terminated string containing the shader code */
static char *LoadShader(const char *path) {
    FILE *f = fopen(path, "rb");
    if (!f) {
        fprintf(stderr, "Failed to load %s: Could not fopen()\n", path);
        return NULL;
    }

    fseek(f, 0, SEEK_END);
    size_t fsz = (size_t)ftell(f);
    fseek(f, 0, SEEK_SET);
    char *buf = malloc(fsz + 1);
    if (!buf) {
        fprintf(stderr, "Failed to load %s: Unable to allocate %zu bytes.\n", path, fsz + 1);
        fclose(f);
        return NULL;
    }
    if (fread(buf, fsz, 1, f) != 1) {
        fprintf(stderr, "Failed to load %s: Unable to read %zu bytes.\n", path, fsz);
        free(buf);
        fclose(f);
        return NULL;
    }
    buf[fsz] = '\0';
    fclose(f);
    return buf;
}

static int CreateGeometryPass(GeometryPass *pass) {
    /* Load vertex & fragment shader */
    char *vert_code = LoadShader("shader/geom_vert.glsl");
    if (!vert_code)
        return 0;
    char *frag_code = LoadShader("shader/geom_frag.glsl");
    if (!frag_code) {
        free(vert_code);
        return 0;
    }

    GLint result;
    GLuint vert = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vert, 1, &vert_code, NULL);
    glCompileShader(vert);
    glGetShaderiv(vert, GL_COMPILE_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetShaderInfoLog(vert, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to compile shader/geom_vert.glsl: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }

    GLuint frag = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(frag, 1, &frag_code, NULL);
    glCompileShader(frag);
    glGetShaderiv(frag, GL_COMPILE_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetShaderInfoLog(frag, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to compile shader/geom_frag.glsl: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }

    pass->prog = glCreateProgram();
    glAttachShader(pass->prog, vert);
    glAttachShader(pass->prog, frag);
    glLinkProgram(pass->prog);
    glGetProgramiv(pass->prog, GL_LINK_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetProgramInfoLog(pass->prog, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to link geometry pass program: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }
    glDeleteShader(vert);
    glDeleteShader(frag);
    free(vert_code);
    free(frag_code);

    glGenBuffers(1, &pass->frame_matrices_ubo);
    glBindBuffer(GL_UNIFORM_BUFFER, pass->frame_matrices_ubo);
    glBufferData(GL_UNIFORM_BUFFER, sizeof(FrameMatrices), NULL, GL_DYNAMIC_DRAW);
    glBindBuffer(GL_UNIFORM_BUFFER, 0);

    return 1;
}

static void CreateHDRBuffer(int w, int h, HDRBuffer *hdr_buf) {
    glGenFramebuffers(1, &hdr_buf->fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, hdr_buf->fbo);

    /* Create the texture objects in one batch */
    glGenTextures(1, &hdr_buf->color);

    /* position buffer */
    glBindTexture(GL_TEXTURE_2D, hdr_buf->color);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F, w, h, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, hdr_buf->color, 0);

    GLuint attachments[1] = { GL_COLOR_ATTACHMENT0 };
    glDrawBuffers(1, attachments);
    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        fprintf(stderr, "ERROR::FRAMEBUFFER:: Framebuffer is not complete!\n");
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

static int CreateLightingPass(LightingPass *pass) {
    /* Load vertex & fragment shader */
    char *vert_code = LoadShader("shader/light_vert.glsl");
    if (!vert_code)
        return 0;
    char *frag_code = LoadShader("shader/light_frag.glsl");
    if (!frag_code) {
        free(vert_code);
        return 0;
    }

    GLint result;
    GLuint vert = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vert, 1, &vert_code, NULL);
    glCompileShader(vert);
    glGetShaderiv(vert, GL_COMPILE_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetShaderInfoLog(vert, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to compile shader/light_vert.glsl: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }

    GLuint frag = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(frag, 1, &frag_code, NULL);
    glCompileShader(frag);
    glGetShaderiv(frag, GL_COMPILE_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetShaderInfoLog(frag, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to compile shader/light_frag.glsl: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }

    pass->prog = glCreateProgram();
    glAttachShader(pass->prog, vert);
    glAttachShader(pass->prog, frag);
    glLinkProgram(pass->prog);
    glGetProgramiv(pass->prog, GL_LINK_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetProgramInfoLog(pass->prog, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to link lighting pass program: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }
    glDeleteShader(vert);
    glDeleteShader(frag);
    free(vert_code);
    free(frag_code);

    glGenBuffers(1, &pass->lights_sbo);
    glBindBuffer(GL_SHADER_STORAGE_BUFFER, pass->lights_sbo);
    glBufferData(GL_SHADER_STORAGE_BUFFER, sizeof(Light) * MAX_LIGHTS, NULL, GL_DYNAMIC_DRAW);
    glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0);

    return 1;
}


static int CreateToneMappingPass(ToneMappingPass *pass) {
    /* Load vertex & fragment shader */
    char *vert_code = LoadShader("shader/tone_map_vert.glsl");
    if (!vert_code)
        return 0;
    char *frag_code = LoadShader("shader/tone_map_frag.glsl");
    if (!frag_code) {
        free(vert_code);
        return 0;
    }

    GLint result;
    GLuint vert = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vert, 1, &vert_code, NULL);
    glCompileShader(vert);
    glGetShaderiv(vert, GL_COMPILE_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetShaderInfoLog(vert, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to compile shader/tone_map_vert.glsl: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }

    GLuint frag = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(frag, 1, &frag_code, NULL);
    glCompileShader(frag);
    glGetShaderiv(frag, GL_COMPILE_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetShaderInfoLog(frag, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to compile shader/tone_map_frag.glsl: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }

    pass->prog = glCreateProgram();
    glAttachShader(pass->prog, vert);
    glAttachShader(pass->prog, frag);
    glLinkProgram(pass->prog);
    glGetProgramiv(pass->prog, GL_LINK_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetProgramInfoLog(pass->prog, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to link lighting pass program: %s\n", info);
        free(vert_code);
        free(frag_code);
        return 0;
    }
    glDeleteShader(vert);
    glDeleteShader(frag);
    free(vert_code);
    free(frag_code);

    return 1;
}

static int CreatePixelClassifierPass(PixelClassifierPass *pass) {
    /* Load vertex & fragment shader */
    char *code = LoadShader("shader/pixel_classifier.glsl");
    if (!code)
        return 0;
    GLint result;
    GLuint comp = glCreateShader(GL_COMPUTE_SHADER);
    glShaderSource(comp, 1, &code, NULL);
    glCompileShader(comp);
    glGetShaderiv(comp, GL_COMPILE_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetShaderInfoLog(comp, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to compile shader/pixel_classifier.glsl: %s\n", info);
        free(code);
        return 0;
    }

    pass->prog = glCreateProgram();
    glAttachShader(pass->prog, comp);
    glLinkProgram(pass->prog);
    glGetProgramiv(pass->prog, GL_LINK_STATUS, &result);
    if (result != GL_TRUE) {
        char info[512];
        glGetProgramInfoLog(pass->prog, sizeof(info), NULL, info);
        fprintf(stderr, "Failed to link lighting pass program: %s\n", info);
        free(code);
        return 0;
    }
    glDeleteShader(comp);
    free(code);

    return 1;
}

static void CreateDifferentiationBuffer(int w, int h, DifferentiationBuffer *diff_buf) {
    glGenTextures(1, &diff_buf->classification);

    /* position buffer */
    glBindTexture(GL_TEXTURE_2D, diff_buf->classification);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_R8UI, w, h, 0, GL_RED_INTEGER, GL_UNSIGNED_BYTE, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glBindTexture(GL_TEXTURE_2D, 0);
}

static void DestroyGBuffer(GBuffer *gbuf) {
    glDeleteFramebuffers(1, &gbuf->fbo);
    glDeleteRenderbuffers(1, &gbuf->depth);
    glDeleteTextures(1, &gbuf->orm);
    glDeleteTextures(1, &gbuf->normal);
    glDeleteTextures(1, &gbuf->albedo);
    glDeleteTextures(1, &gbuf->position);
    memset(gbuf, 0, sizeof(*gbuf));
}

static void DestroyGeometryPass(GeometryPass *pass) {
    glDeleteProgram(pass->prog);
    memset(pass, 0, sizeof(*pass));
}

static void DestroyHDRBuffer(HDRBuffer *hdr_buf) {
    glDeleteFramebuffers(1, &hdr_buf->fbo);
    glDeleteTextures(1, &hdr_buf->color);
    memset(hdr_buf, 0, sizeof(*hdr_buf));
}

static void DestroyLightingPass(LightingPass *pass) {
    glDeleteProgram(pass->prog);
    glDeleteBuffers(1, &pass->lights_sbo);
    memset(pass, 0, sizeof(*pass));
}

static void DestroyToneMappingPass(ToneMappingPass *pass) {
    glDeleteProgram(pass->prog);
    memset(pass, 0, sizeof(*pass));
}

static void DestroyPixelClassificationPass(PixelClassifierPass *pass) {
    glDeleteProgram(pass->prog);
    memset(pass, 0, sizeof(*pass));
}

static void DestroyDifferentiationBuffer(DifferentiationBuffer *diff_buf) {
    glDeleteTextures(1, &diff_buf->classification);
    memset(diff_buf, 0, sizeof(*diff_buf));
}

/* Create the deferred renderer.
 * w and h give the size of the gbuffer in pixels. */
DeferredRenderer *CreateDeferredRenderer(int w, int h, float near, float far) {
    DeferredRenderer *renderer = malloc(sizeof(DeferredRenderer));
    if (!renderer)
        return NULL;
    memset(renderer, 0, sizeof(*renderer));

    CreateGBuffer(w, h, &renderer->gbuf);
    if (!CreateGeometryPass(&renderer->geometry_pass)) {
        DestroyGBuffer(&renderer->gbuf);
        free(renderer);
        return NULL;
    }

    CreateHDRBuffer(w, h, &renderer->hdr_buf);
    if (!CreateLightingPass(&renderer->lighting_pass)) {
        DestroyHDRBuffer(&renderer->hdr_buf);
        DestroyGeometryPass(&renderer->geometry_pass);
        DestroyGBuffer(&renderer->gbuf);
        free(renderer);
        return NULL;
    }

    if (!CreateToneMappingPass(&renderer->tone_mapping_pass)) {
        DestroyLightingPass(&renderer->lighting_pass);
        DestroyHDRBuffer(&renderer->hdr_buf);
        DestroyGeometryPass(&renderer->geometry_pass);
        DestroyGBuffer(&renderer->gbuf);
        free(renderer);
        return NULL;
    }

    if (!CreatePixelClassifierPass(&renderer->classifier_pass)) {
        DestroyToneMappingPass(&renderer->tone_mapping_pass);
        DestroyLightingPass(&renderer->lighting_pass);
        DestroyHDRBuffer(&renderer->hdr_buf);
        DestroyGeometryPass(&renderer->geometry_pass);
        DestroyGBuffer(&renderer->gbuf);
        free(renderer);
    }
    CreateDifferentiationBuffer(w, h, &renderer->diff_buffer);

    glGenVertexArrays(1, &renderer->full_screen_triangle_vao);
    glBindVertexArray(renderer->full_screen_triangle_vao);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindVertexArray(0);

    renderer->light_count = 0;
    renderer->light_ssbo_needs_update = 0;
    renderer->projection = Perspective(DegToRad(45.f), (float)w / (float)h, near, far);

    renderer->width = w;
    renderer->height = h;

    return renderer;
}

/* Free all resources allocated for the deferred renderer. */
void DestroyDeferredRenderer(DeferredRenderer *ren) {
    DestroyDifferentiationBuffer(&ren->diff_buffer);
    DestroyPixelClassificationPass(&ren->classifier_pass);
    DestroyToneMappingPass(&ren->tone_mapping_pass);
    DestroyLightingPass(&ren->lighting_pass);
    DestroyHDRBuffer(&ren->hdr_buf);
    DestroyGeometryPass(&ren->geometry_pass);
    DestroyGBuffer(&ren->gbuf);
    free(ren);
}

void BeginFrame(DeferredRenderer *renderer, V3 eye_pos, V3 center, V3 up, V3 ambient_color) {
    renderer->eye_pos = eye_pos;
    renderer->center = center;
    renderer->up = up;
    renderer->ambient_color = ambient_color;

    if (renderer->light_ssbo_needs_update) {
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, renderer->lighting_pass.lights_sbo);
        glBufferSubData(GL_SHADER_STORAGE_BUFFER, 0, sizeof(Light) * renderer->light_count, renderer->lights);
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0);
        renderer->light_ssbo_needs_update = 0;
    }
    glBindFramebuffer(GL_FRAMEBUFFER, renderer->gbuf.fbo);
    glClearColor(0.f, 0.f, 0.f, 1.f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

}

void AddLight(DeferredRenderer *renderer, V3 center, V3 color, float radius) {
    if (renderer->light_count == MAX_LIGHTS)
        return;
    Light l;
    l.position_radius = (V4){ .x = center.x, .y = center.y, .z = center.z, .w = radius };
    l.color = (V4){ color.x, color.y, color.z, 0.f };
    renderer->lights[renderer->light_count++] = l;
    renderer->light_ssbo_needs_update = 1;
}

void RenderModel(DeferredRenderer *renderer, const Model *model, M4 transform) {
    glBindFramebuffer(GL_FRAMEBUFFER, renderer->gbuf.fbo);
    glUseProgram(renderer->geometry_pass.prog);
    glEnable(GL_DEPTH_TEST);

    /* View & Projection */
    FrameMatrices frame_matrices;
    frame_matrices.projection = renderer->projection;
    frame_matrices.view = LookAt(renderer->eye_pos, renderer->center, renderer->up);

    glBindBuffer(GL_UNIFORM_BUFFER, renderer->geometry_pass.frame_matrices_ubo);
    glBufferSubData(GL_UNIFORM_BUFFER, 0, sizeof(frame_matrices), &frame_matrices);
    glBindBuffer(GL_UNIFORM_BUFFER, 0);
    glBindBufferBase(GL_UNIFORM_BUFFER, GEOMETRY_PASS_FRAME_MATRICES_BINDING, renderer->geometry_pass.frame_matrices_ubo);

    glUniform1i(GEOMETRY_PASS_ALBEDO_LOCATION, 0);
    glUniform1i(GEOMETRY_PASS_NORMAL_LOCATION, 1);
    glUniform1i(GEOMETRY_PASS_ORM_LOCATION, 2);

    for (unsigned int i = 0; i < model->mesh_count; ++i) {
        glBindVertexArray(model->meshes[i].vao);
        glUniformMatrix4fv(GEOMETRY_PASS_MODEL_LOCATION, 1, GL_FALSE, &transform.e[0]);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, model->meshes[i].material->albedo->tex);
        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, model->meshes[i].material->normal->tex);
        glActiveTexture(GL_TEXTURE2);
        glBindTexture(GL_TEXTURE_2D, model->meshes[i].material->orm->tex);
        glDrawElements(GL_TRIANGLES, model->meshes[i].index_count, GL_UNSIGNED_INT, NULL);
        glBindVertexArray(0);
    }
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void ResolveGBuffer(DeferredRenderer *renderer) {
    glBindFramebuffer(GL_FRAMEBUFFER, renderer->hdr_buf.fbo);
    glUseProgram(renderer->lighting_pass.prog);
    glDisable(GL_DEPTH_TEST);

    glClearColor(0.f, 0.f, 0.f, 1.f);
    glClear(GL_COLOR_BUFFER_BIT);

    glUniform1ui(LIGHTING_PASS_LIGHT_COUNT_LOCATION, renderer->light_count);
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, LIGHTING_PASS_LIGHTS_BINDING, renderer->lighting_pass.lights_sbo);

    glUniform3f(LIGHTING_PASS_AMBIENT_COLOR_LOCATION, renderer->ambient_color.x, renderer->ambient_color.y, renderer->ambient_color.z);

    glUniform1i(LIGHTING_PASS_POSITION_LOCATION, 0);
    glUniform1i(LIGHTING_PASS_ALBEDO_LOCATION, 1);
    glUniform1i(LIGHTING_PASS_NORMAL_LOCATION, 2);
    glUniform1i(LIGHTING_PASS_ORM_LOCATION, 3);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, renderer->gbuf.position);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, renderer->gbuf.albedo);
    glActiveTexture(GL_TEXTURE2);
    glBindTexture(GL_TEXTURE_2D, renderer->gbuf.normal);
    glActiveTexture(GL_TEXTURE3);
    glBindTexture(GL_TEXTURE_2D, renderer->gbuf.orm);

    glBindVertexArray(renderer->full_screen_triangle_vao);
    glDrawArrays(GL_TRIANGLES, 0, 3);
    glBindVertexArray(0);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);

}

void ToneMapping(DeferredRenderer *renderer) {
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glUseProgram(renderer->tone_mapping_pass.prog);
    glDisable(GL_DEPTH_TEST);

    glClearColor(0.f, 0.f, 0.f, 1.f);
    glClear(GL_COLOR_BUFFER_BIT);

    glUniform1i(TONE_MAPPING_PASS_COLOR_LOCATION, 0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, renderer->hdr_buf.color);

    glBindVertexArray(renderer->full_screen_triangle_vao);
    glDrawArrays(GL_TRIANGLES, 0, 3);
    glBindVertexArray(0);
}

void Differentiate(DeferredRenderer *renderer) {
    glUseProgram(renderer->classifier_pass.prog);
    GLuint groups_x = (GLuint)(renderer->width + 15) / 16;
    GLuint groups_y = (GLuint)(renderer->height + 15) / 16;

    glBindImageTexture(PIXEL_CLASSIFIER_PASS_NORMAL_BINDING, renderer->gbuf.normal, 0, GL_FALSE, 0, GL_READ_ONLY, GL_RGBA16F);
    glBindImageTexture(PIXEL_CLASSIFIER_PASS_CLASSIFICATION_BINDING, renderer->diff_buffer.classification, 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R8UI);

    glDispatchCompute(groups_x, groups_y, 1);
}