#include <glad/gl.h>
#include <GLFW/glfw3.h>
#include <stdio.h>

#include "diffren.h"

void APIENTRY glDebugOutput(GLenum source,
    GLenum type,
    unsigned int id,
    GLenum severity,
    GLsizei length,
    const char *message,
    const void *userParam)
{
    UNUSED(length);
    UNUSED(userParam);

    // ignore non-significant error/warning codes
    if (id == 131169 || id == 131185 || id == 131218 || id == 131204)
        return;

    fprintf(stderr, "---------------\n");
    fprintf(stderr, "Debug message (%u): %s\n", id, message);
    
    switch (source)
    {
    case GL_DEBUG_SOURCE_API:             fprintf(stderr, "Source: API"); break;
    case GL_DEBUG_SOURCE_WINDOW_SYSTEM:   fprintf(stderr, "Source: Window System"); break;
    case GL_DEBUG_SOURCE_SHADER_COMPILER: fprintf(stderr, "Source: Shader Compiler"); break;
    case GL_DEBUG_SOURCE_THIRD_PARTY:     fprintf(stderr, "Source: Third Party"); break;
    case GL_DEBUG_SOURCE_APPLICATION:     fprintf(stderr, "Source: Application"); break;
    case GL_DEBUG_SOURCE_OTHER:           fprintf(stderr, "Source: Other"); break;
    }
    fprintf(stderr, "\n");

    switch (type)
    {
    case GL_DEBUG_TYPE_ERROR:               fprintf(stderr, "Type: Error"); break;
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR: fprintf(stderr, "Type: Deprecated Behaviour"); break;
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:  fprintf(stderr, "Type: Undefined Behaviour"); break;
    case GL_DEBUG_TYPE_PORTABILITY:         fprintf(stderr, "Type: Portability"); break;
    case GL_DEBUG_TYPE_PERFORMANCE:         fprintf(stderr, "Type: Performance"); break;
    case GL_DEBUG_TYPE_MARKER:              fprintf(stderr, "Type: Marker"); break;
    case GL_DEBUG_TYPE_PUSH_GROUP:          fprintf(stderr, "Type: Push Group"); break;
    case GL_DEBUG_TYPE_POP_GROUP:           fprintf(stderr, "Type: Pop Group"); break;
    case GL_DEBUG_TYPE_OTHER:               fprintf(stderr, "Type: Other"); break;
    }
    fprintf(stderr, "\n");

    switch (severity)
    {
    case GL_DEBUG_SEVERITY_HIGH:         fprintf(stderr, "Severity: high"); break;
    case GL_DEBUG_SEVERITY_MEDIUM:       fprintf(stderr, "Severity: medium"); break;
    case GL_DEBUG_SEVERITY_LOW:          fprintf(stderr, "Severity: low"); break;
    case GL_DEBUG_SEVERITY_NOTIFICATION: fprintf(stderr, "Severity: notification"); break;
    }
    fprintf(stderr, "\n\n");
}

static void SetupDebugOutput() {
    int flags; 
    glGetIntegerv(GL_CONTEXT_FLAGS, &flags);
    if (flags & GL_CONTEXT_FLAG_DEBUG_BIT) {
        glEnable(GL_DEBUG_OUTPUT);
        glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
        glDebugMessageCallback(glDebugOutput, NULL);
        glDebugMessageControl(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, NULL, GL_TRUE);
    }
}

int main(int argc, char **argv) {
    UNUSED(argc);
    UNUSED(argv);

    if (!glfwInit())
        return 1;

    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 6);
#ifndef NDEBUG
    glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GLFW_TRUE);
#endif
    GLFWwindow *window = glfwCreateWindow(1024, 768, "diffren", NULL, NULL);
    if (!window) {
        glfwTerminate();
        return 1;
    }
    glfwMakeContextCurrent(window);
    if (!gladLoadGL((GLADloadfunc)glfwGetProcAddress)) {
        glfwDestroyWindow(window);
        glfwTerminate();
        return 1;
    }
#ifndef NDEBUG
    SetupDebugOutput();
#endif
    /* Vsync */
    glfwSwapInterval(1);

    int fbuf_w, fbuf_h;
    glfwGetFramebufferSize(window, &fbuf_w, &fbuf_h);
    DeferredRenderer *renderer = CreateDeferredRenderer(fbuf_w, fbuf_h, 0.1f, 100.f);

    Model *waterbottle = ImportModel("model/WaterBottle.gltf");

    AddLight(renderer, (V3) { -1.f, -1.f, -1.f, }, (V3) { 12.f, 12.f, 4.f }, 3.f);
    AddLight(renderer, (V3) { 1.f, 1.f, -1.f, }, (V3) { 2.f, 2.f, 20.f }, 3.f);

    float angle = 0.f;

    while (!glfwWindowShouldClose(window)) {
        glfwPollEvents();

        BeginFrame(renderer, (V3) { -1.f, 0.f, 0.f }, (V3) { 0.f, 0.f, 0.f }, (V3) { 0.f, 1.f, 0.f }, (V3) { 0.3f, 0.3f, 0.3f });

        M4 transform = M4Rotate(1.f, 0.f, 1.f, DegToRad(angle));
        RenderModel(renderer, waterbottle, transform);

        transform = M4Mul(M4Translate(0.f, 0.f, 0.2f), M4Rotate(0.f, 1.f, 0.f, DegToRad(angle)));
        RenderModel(renderer, waterbottle, transform);

        angle += 0.5f;

        ResolveGBuffer(renderer);
        ToneMapping(renderer);
        Differentiate(renderer);

        glfwSwapBuffers(window);
    }

    DestroyDeferredRenderer(renderer);
    glfwDestroyWindow(window);
    glfwTerminate();
    return 0;
}
