#include "diffren_math.h"

#include <math.h>
#include <string.h>

static float V4MulInner(V4 a, V4 b) {
    float p = 0.f;
    int i;
    for (i = 0; i < 4; ++i)
        p += b.e[i] * a.e[i];
    return p;
}

static float V3Length(V3 v) {
    return sqrtf(v.x * v.x + v.y * v.y + v.z * v.z);
}

V3 V3Norm(V3 v) {
    float len = sqrtf(v.x * v.x + v.y * v.y + v.z * v.z);
    V3 n = { .x = v.x / len, .y = v.y / len, .z = v.z / len };
    return n;
}

static V4 Mat4Row(const M4 *m, int i)
{
    V4 r;
    int k;
    for (k = 0; k < 4; ++k)
        r.e[k] = m->columns[k].e[i];
    return r;
}

static void M4TranslateInPlace(M4 *m, float x, float y, float z)
{
    V4 t = { x, y, z, 0 };
    V4 r;
    int i;
    for (i = 0; i < 4; ++i) {
        r = Mat4Row(m, i);
        m->columns[3].e[i] += V4MulInner(r, t);
    }
}

M4 M4Add(M4 a, M4 b) {
    M4 s;
    for (int i = 0; i < 4 * 4; ++i)
        s.e[i] = a.e[i] + b.e[i];
    return s;
}

M4 M4Sub(M4 a, M4 b) {
    M4 s;
    for (int i = 0; i < 4 * 4; ++i)
        s.e[i] = a.e[i] - b.e[i];
    return s;
}

M4 M4Mul(M4 a, M4 b)
{
    M4 temp;
    int k, r, c;
    for (c = 0; c < 4; ++c) {
        for (r = 0; r < 4; ++r) {
            temp.columns[c].e[r] = 0.f;
            for (k = 0; k < 4; ++k)
                temp.columns[c].e[r] += a.columns[k].e[r] * b.columns[c].e[k];
        }
    }
    return temp;
}

M4 M4FromV3Outer(V3 a, V3 b)
{
    M4 M;
    int i, j;
    for (i = 0; i < 4; ++i) for (j = 0; j < 4; ++j)
        M.columns[i].e[j] = i < 3 && j < 3 ? a.e[i] * b.e[j] : 0.f;
    return M;
}

M4 LookAt(V3 eye, V3 center, V3 up) {
    M4 m;
    V3 f = V3Sub(center, eye);
    f = V3Norm(f);

    V3 s = V3Cross(f, up);
    s = V3Norm(s);

    V3 t = V3Cross(s, f);

    m.columns[0].e[0] = s.x;
    m.columns[0].e[1] = t.x;
    m.columns[0].e[2] = -f.x;
    m.columns[0].e[3] = 0.f;

    m.columns[1].e[0] = s.y;
    m.columns[1].e[1] = t.y;
    m.columns[1].e[2] = -f.y;
    m.columns[1].e[3] = 0.f;

    m.columns[2].e[0] = s.z;
    m.columns[2].e[1] = t.z;
    m.columns[2].e[2] = -f.z;
    m.columns[2].e[3] = 0.f;

    m.columns[3].e[0] = 0.f;
    m.columns[3].e[1] = 0.f;
    m.columns[3].e[2] = 0.f;
    m.columns[3].e[3] = 1.f;

    M4TranslateInPlace(&m, -eye.x, -eye.y, -eye.z);
    return m;
}

M4 Perspective(float y_fov, float aspect, float n, float f)
{
    M4 m;
    float a = 1.f / tanf(y_fov / 2.f);

    m.columns[0].e[0] = a / aspect;
    m.columns[0].e[1] = 0.f;
    m.columns[0].e[2] = 0.f;
    m.columns[0].e[3] = 0.f;

    m.columns[1].e[0] = 0.f;
    m.columns[1].e[1] = a;
    m.columns[1].e[2] = 0.f;
    m.columns[1].e[3] = 0.f;

    m.columns[2].e[0] = 0.f;
    m.columns[2].e[1] = 0.f;
    m.columns[2].e[2] = -((f + n) / (f - n));
    m.columns[2].e[3] = -1.f;

    m.columns[3].e[0] = 0.f;
    m.columns[3].e[1] = 0.f;
    m.columns[3].e[2] = -((2.f * f * n) / (f - n));
    m.columns[3].e[3] = 0.f;
    
    return m;
}

M4 M4MulScalar(M4 m, float s) {
    M4 t;
    for (int i = 0; i < 4 * 4; ++i)
        t.e[i] = m.e[i] * s;
    return t;
}

M4 M4Identity(void) {
    M4 m;
    memset(&m, 0, sizeof(m));
    m.columns[0].e[0] = 1.f;
    m.columns[1].e[1] = 1.f;
    m.columns[2].e[2] = 1.f;
    m.columns[3].e[3] = 1.f;
    return m;
}

M4 M4Translate(float x, float y, float z) {
    M4 m = M4Identity();
    m.columns[3].x = x;
    m.columns[3].y = y;
    m.columns[3].z = z;
    return m;
}

M4 M4Rotate(float x, float y, float z, float angle) {
    float s = sinf(angle);
    float c = cosf(angle);
    V3 u = { x, y, z };

    if (V3Length(u) > 1e-4) {
        u = V3Norm(u);
        M4 T = M4FromV3Outer(u, u);

        M4 S = {
            .columns[0] = {    0,  u.z, -u.y,  0},
            .columns[1] = {-u.z,     0,  u.x,  0},
            .columns[2] = { u.y,  -u.x,     0, 0},
            .columns[3] = {    0,    0,     0, 0}
        };
        S = M4MulScalar(S, s);

        M4 C = M4Identity();
        C = M4Sub(C, T);

        C = M4MulScalar(C, c);

        T = M4Add(T, C);
        T = M4Add(T, S);

        T.columns[3].e[3] = 1.f;
        return T;
    }
    else {
        return M4Identity();
    }
}