#include <assimp/cimport.h>
#include <assimp/scene.h>
#include <assimp/postprocess.h>
#include <stdlib.h>
#include <stdio.h>

#include "diffren.h"
#include "diffren_model.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

/* TODO(Kevin): Handle the node structure.
 * Handle multiple meshes using the same texture.
 * Turn allocations into flat preallocated buffers
 */

static int ProcessTexture(struct aiMaterial *mat, enum aiTextureType type, Model *model) {
    if (aiGetMaterialTextureCount(mat, type) == 0)
        return 0;

    struct aiString path;
    if (aiGetMaterialTexture(mat, type, 0, &path, NULL, NULL, NULL, NULL, NULL, NULL) != aiReturn_SUCCESS)
        return 0;

    Texture tex;
    tex.name = malloc(path.length + 1);
    if (!tex.name)
        return 0;
    memcpy(tex.name, path.data, path.length + 1);

    char full_path[260];
    snprintf(full_path, 259, "%s%s", model->directory, path.data);
    full_path[259] = '\0';

    stbi_uc *pixels = stbi_load(full_path, &tex.width, &tex.height, &tex.channels, 0);
    if (!pixels) {
        free(tex.name);
        return 0;
    }
    glGenTextures(1, &tex.tex);
    glBindTexture(GL_TEXTURE_2D, tex.tex);
    GLenum internal_format, format;
    if (tex.channels == 3) {
        internal_format = GL_RGB8;
        format = GL_RGB;
    }
    else if (tex.channels == 4) {
        internal_format = GL_RGBA8;
        format = GL_RGBA;
    }
    else if (tex.channels == 2) {
        internal_format = GL_RG8;
        format = GL_RG;
    }
    else if (tex.channels == 1) {
        internal_format = GL_R8;
        format = GL_RED;
    }
    else {
        free(tex.name);
        stbi_image_free(pixels);
        return 0;
    }
    glTexImage2D(GL_TEXTURE_2D, 0, internal_format, tex.width, tex.height, 0, format, GL_UNSIGNED_BYTE, pixels);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    stbi_image_free(pixels);

    model->textures[model->texture_count++] = tex;
    return 1;
}

static int ProcessMaterial(struct aiMaterial *in_material, Model *model) {
    Material mat;
    if (ProcessTexture(in_material, aiTextureType_BASE_COLOR, model))
        mat.albedo = &model->textures[model->texture_count - 1];
    else
        mat.albedo = NULL;
    if (ProcessTexture(in_material, aiTextureType_NORMALS, model))
        mat.normal = &model->textures[model->texture_count - 1];
    else
        mat.normal = NULL;
    if (ProcessTexture(in_material, aiTextureType_DIFFUSE_ROUGHNESS, model))
        mat.orm = &model->textures[model->texture_count - 1];
    else
        mat.orm = NULL;
    model->materials[model->material_count++] = mat;
    return 1;
}

static int ProcessMesh(struct aiMesh *in_mesh, const struct aiScene *scene, Model *model) {
    Mesh m;
    m.vertices = malloc(in_mesh->mNumVertices * sizeof(Vertex));
    if (!m.vertices)
        return 0;
    m.indices = malloc(in_mesh->mNumFaces * 3 * sizeof(uint32_t));
    if (!m.indices) {
        free(m.vertices);
        return 0;
    }
    m.vertex_count = in_mesh->mNumVertices;
    m.index_count = 0;

    for (unsigned int i = 0; i < in_mesh->mNumVertices; ++i) {
        Vertex v;
        v.position.x = in_mesh->mVertices[i].x;
        v.position.y = in_mesh->mVertices[i].y;
        v.position.z = in_mesh->mVertices[i].z;

        v.normal.x = in_mesh->mNormals[i].x;
        v.normal.y = in_mesh->mNormals[i].y;
        v.normal.z = in_mesh->mNormals[i].z;

        v.tangent.x = in_mesh->mTangents[i].x;
        v.tangent.y = in_mesh->mTangents[i].y;
        v.tangent.z = in_mesh->mTangents[i].z;

        if (in_mesh->mTextureCoords[0]) {
            v.texcoord.x = in_mesh->mTextureCoords[0][i].x;
            v.texcoord.y = in_mesh->mTextureCoords[0][i].y;
        }
        else {
            v.texcoord.x = 0.f;
            v.texcoord.y = 0.f;
        }

        m.vertices[i] = v;
    }

    for (unsigned int i = 0; i < in_mesh->mNumFaces; ++i) {
        struct aiFace face = in_mesh->mFaces[i];
        for (unsigned int j = 0; j < face.mNumIndices; ++j) {
            m.indices[m.index_count++] = face.mIndices[j];
        }
    }

    if (in_mesh->mMaterialIndex >= 0) {
        if (!ProcessMaterial(scene->mMaterials[in_mesh->mMaterialIndex], model)) {
            free(m.vertices);
            free(m.indices);
            return 0;
        }
        m.material = &model->materials[model->material_count - 1];
    }
    else {
        m.material = NULL;
    }

    glGenBuffers(1, &m.vbo);
    glGenBuffers(1, &m.ebo);
    glGenVertexArrays(1, &m.vao);
    
    glBindVertexArray(m.vao);
    glBindBuffer(GL_ARRAY_BUFFER, m.vbo);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vertex) * m.vertex_count, m.vertices, GL_STATIC_DRAW);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m.ebo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(uint32_t) * m.index_count, m.indices, GL_STATIC_DRAW);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, position));
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, normal));
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tangent));
    glVertexAttribPointer(3, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, texcoord));
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    glEnableVertexAttribArray(2);
    glEnableVertexAttribArray(3);

    glBindVertexArray(0);

    model->meshes[model->mesh_count++] = m;
    return 1;
}

static int ProcessNode(struct aiNode *node, const struct aiScene *scene, Model *model) {
    for (unsigned int i = 0; i < node->mNumMeshes; ++i) {
        struct aiMesh *mesh = scene->mMeshes[node->mMeshes[i]];
        if (!ProcessMesh(mesh, scene, model))
            return 0;
    }

    for (unsigned int i = 0; i < node->mNumChildren; ++i) {
        if (!ProcessNode(node->mChildren[i], scene, model))
            return 0;
    }
    return 1;
}

/* Import a 3d model from the given file */
Model *ImportModel(const char *file) {
    Model *model = malloc(sizeof(*model));
    if (!model) {
        fprintf(stderr, "Could not import %s: Could not allocate storage.\n", file);
        return NULL;
    }

    /* Determine directory */
    model->directory = malloc(strlen(file) + 1);
    if (!model->directory) {
        fprintf(stderr, "Could not import %s: Could not allocate directory path storage.\n", file);
        free(model);
        return NULL;
    }
    memcpy(model->directory, file, strlen(file) + 1);
    for (int i = (int)strlen(file); i >= 0; --i) {
        if (model->directory[i] == '/' || model->directory[i] == '\\') {
            model->directory[i + 1] = '\0';
            break;
        }
    }

    const struct aiScene *scene = aiImportFile(file,
        aiProcess_CalcTangentSpace |
        aiProcess_Triangulate |
        aiProcess_JoinIdenticalVertices |
        aiProcess_SortByPType |
        aiProcess_OptimizeMeshes |
        aiProcess_FlipUVs |
        aiProcess_GenNormals);
    if (!scene || (scene->mFlags & AI_SCENE_FLAGS_INCOMPLETE) != 0 || !scene->mRootNode) {
        fprintf(stderr, "Could not import %s: %s", file, aiGetErrorString());
        free(model);
        return NULL;
    }

    model->meshes = malloc(sizeof(Mesh) * scene->mNumMeshes);
    if (!model->meshes) {
        fprintf(stderr, "Could not import %s: Could not allocate mesh storage.\n", file);
        free(model);
        aiReleaseImport(scene);
        return NULL;
    }
    model->mesh_count = 0;

    model->materials = malloc(sizeof(Material) * scene->mNumMaterials);
    if (!model->materials) {
        fprintf(stderr, "Could not import %s: Could not allocate material storage.\n", file);
        free(model->meshes);
        free(model);
        aiReleaseImport(scene);
        return NULL;
    }
    model->material_count = 0;

    model->textures = malloc(sizeof(Texture) * scene->mNumMaterials * 3);
    if (!model->textures) {
        fprintf(stderr, "Could not import %s: Could not allocate texture storage.\n", file);
        free(model->textures);
        free(model->meshes);
        free(model);
        aiReleaseImport(scene);
        return NULL;
    }
    model->texture_count = 0;

    if (!ProcessNode(scene->mRootNode, scene, model)) {
        DestroyModel(model);
        model = NULL;
    }

    aiReleaseImport(scene);
    return model;
}

/* Destroy an imported model */
void DestroyModel(Model *model) {
    free(model->directory);
    for (unsigned int i = 0; i < model->texture_count; ++i) {
        free(model->textures[i].name);
        glDeleteTextures(1, &model->textures[i].tex);
    }
    free(model->textures);
    free(model->materials);
    for (unsigned int i = 0; i < model->mesh_count; ++i) {
        free(model->meshes[i].vertices);
        free(model->meshes[i].indices);
        glDeleteBuffers(1, &model->meshes[i].vbo);
        glDeleteBuffers(1, &model->meshes[i].ebo);
        glDeleteVertexArrays(1, &model->meshes[i].vbo);
    }
    free(model->meshes);
    free(model);
}